## ----setup, include = FALSE---------------------------------------------------
BiocStyle::markdown()

dir.create("tempOutput")

set.seed(1)

## ----load_packages, message=FALSE, warning=FALSE------------------------------
library(SummarizedExperiment)
library(ReducedExperiment)
library(ggplot2)

theme_set(theme_bw())

## ----install_bioc, eval=FALSE-------------------------------------------------
# if (!requireNamespace("BiocManager", quietly = TRUE)) {
#     install.packages("BiocManager")
# }
# 
# BiocManager::install("ReducedExperiment")

## ----install_gh, eval=FALSE---------------------------------------------------
# devtools::install_github("jackgisby/ReducedExperiment")

## ----se_vignette, eval=FALSE--------------------------------------------------
# vignette("SummarizedExperiment", package = "SummarizedExperiment")

## ----load_data----------------------------------------------------------------
wave1_se <- readRDS(system.file(
    "extdata", "wave1.rds",
    package = "ReducedExperiment"
))
wave2_se <- readRDS(system.file(
    "extdata", "wave2.rds",
    package = "ReducedExperiment"
))

wave1_se

## ----estimate_stability, eval=FALSE, message=FALSE, warning=FALSE, include=TRUE----
# stability_res <- estimate_stability(
#     wave1_se,
#     n_runs = 100,
#     min_components = 10, max_components = 60, by = 2,
#     mean_stability_threshold = 0.85,
#     verbose = FALSE
# )
# 
# stability_plot <- plot_stability(stability_res)
# print(stability_plot$combined_plot)

## ----run_factor_analysis------------------------------------------------------
wave1_fe <- estimate_factors(
    wave1_se,
    nc = 35,
    use_stability = TRUE,
    n_runs = 30,
    stability_threshold = 0.25
)

wave1_fe

## ----get_factor_reduced-------------------------------------------------------
# get reduced data for first 5 samples and factors
reduced(wave1_fe)[1:5, 1:5]

## ----get_factor_loadings------------------------------------------------------
# get loadings for first 5 genes and factors
loadings(wave1_fe)[1:5, 1:5]

## ----slice_fe-----------------------------------------------------------------
dim(wave1_fe[1:20, 1:10, 1:5])

## ----assess_threshold---------------------------------------------------------
WGCNA::disableWGCNAThreads()

wave1_fit_indices <- assess_soft_threshold(wave1_se)

wave1_estimated_power <- 
    wave1_fit_indices$Power[wave1_fit_indices$estimated_power]
print(paste0("Estimated power: ", wave1_estimated_power))

## -----------------------------------------------------------------------------
wave1_me <- identify_modules(
    wave1_se,
    verbose = 0,
    power = wave1_estimated_power,
    minModuleSize = 10
)

plotDendro(wave1_me)

## ----print_me-----------------------------------------------------------------
wave1_me

## ----get_module_reduced-------------------------------------------------------
# get reduced data for first 5 samples and factors
reduced(wave1_me)[1:5, 1:5]

## ----get_module_assignments---------------------------------------------------
# get assignments for first 5 genes
assignments(wave1_me)[1:5]

## ----get_module_loadings------------------------------------------------------
# get loadings for first 5 genes
loadings(wave1_me)[1:5]

## ----show_coldata-------------------------------------------------------------
selected_cols <- c(
    "sample_id", 
    "individual_id", 
    "case_control", 
    "sex", 
    "calc_age"
)

colData(wave1_se)[, selected_cols]

## ----associate_components-----------------------------------------------------
f <- "~ case_control + sex + calc_age"

associations_me <- associate_components(wave1_me, method = "lm", formula = f)
associations_fe <- associate_components(wave1_fe, method = "lm", formula = f)

## ----associate_table----------------------------------------------------------
associations_fe$summaries[
    associations_fe$summaries$term == "case_controlPOSITIVE" &
        associations_fe$summaries$adj_pvalue < 0.05,
    c("term", "component", "estimate", "stderr", "adj_pvalue")
]

## ----plot_factor_associations-------------------------------------------------
ggplot(
    associations_fe$summaries[
        associations_fe$summaries$term == "case_controlPOSITIVE",
    ],
    aes(-log10(adj_pvalue),
        reorder(component, -log10(adj_pvalue)),
        fill = estimate
    )
) +
    geom_point(pch = 21, size = 3) +
    xlab("-log10(p-value)") +
    ylab("Factor name (ordered by p-value)") +
    geom_vline(xintercept = -log10(0.05)) +
    scale_fill_gradient2(low = "#3A3A98", high = "#832424")

## ----plot_module_associations-------------------------------------------------
ggplot(
    associations_me$summaries[
        associations_me$summaries$term == "case_controlPOSITIVE",
    ],
    aes(-log10(adj_pvalue),
        reorder(component, -log10(adj_pvalue)),
        fill = estimate
    )
) +
    geom_point(pch = 21, size = 3) +
    xlab("-log10(p-value)") +
    ylab("Module name (ordered by p-value)") +
    geom_vline(xintercept = -log10(0.05)) +
    scale_fill_gradient2(low = "#3A3A98", high = "#832424")

## ----module_centrality--------------------------------------------------------
module_centrality <- getCentrality(wave1_me)
head(module_centrality)

## ----factor_alignments--------------------------------------------------------
aligned_features <- getAlignedFeatures(wave1_fe, format = "data.frame")
head(aligned_features[, c("component", "feature", "value")])

## ----run_enrich, message=FALSE, warning=FALSE---------------------------------

TERM2GENE <- get_msigdb_t2g()

factor_5_enrich <- runEnrich(
    wave1_fe[, , "factor_5"], 
    method = "overrepresentation", 
    as_dataframe = TRUE,
    TERM2GENE = TERM2GENE
)

module_1_enrich <- runEnrich(
    wave1_me[, , "module_1"], 
    method = "overrepresentation", 
    as_dataframe = TRUE,
    TERM2GENE = TERM2GENE
)

## ----example_factor_enrichment------------------------------------------------
rownames(factor_5_enrich) <- NULL
factor_5_enrich[, c("ID", "pvalue", "p.adjust", "geneID")]

## ----example_module_enrichment------------------------------------------------

ggplot(
    module_1_enrich[1:15, ],
    aes(-log10(p.adjust), reorder(substr(ID, 1, 45), -log10(p.adjust)))
) +
    geom_point(pch = 21, size = 3) +
    xlab("-log10(p-value)") +
    ylab("Pathway name (ordered by p-value)") +
    expand_limits(x = 0) +
    geom_vline(xintercept = -log10(0.05))

## ----project_data-------------------------------------------------------------
wave2_fe <- projectData(wave1_fe, wave2_se)
wave2_fe

## ----reprint_original---------------------------------------------------------
wave1_fe

## ----replicating_associations-------------------------------------------------
replication_fe <- associate_components(
    wave2_fe, 
    method = "lmer", 
    formula = paste0(f, " + (1|individual_id)")
)

## ----plotting_replicated_associations-----------------------------------------
ggplot(
    replication_fe$summaries[
        replication_fe$summaries$term == "case_controlPOSITIVE",
    ],
    aes(
        -log10(adj_pvalue),
        reorder(component, -log10(adj_pvalue)),
        fill = estimate
    )
) +
    geom_point(pch = 21, size = 3) +
    xlab("-log10(p-value)") +
    ylab("Module nafe (ordered by p-value)") +
    geom_vline(xintercept = -log10(0.05)) +
    scale_fill_gradient2(low = "#3A3A98", high = "#832424") +
    expand_limits(x = 0)

## ----correlation_with_projection----------------------------------------------
original_estimates <- associations_fe$summaries$estimate[
    associations_fe$summaries$term == "case_controlPOSITIVE"
]
replication_estimates <- replication_fe$summaries$estimate[
    replication_fe$summaries$term == "case_controlPOSITIVE"
]

cor.test(original_estimates, replication_estimates)

## ----calc_eigengenes_new_data-------------------------------------------------
wave2_me <- calcEigengenes(wave1_me, wave2_se)

replication_me <- associate_components(
    wave2_me, 
    method = "lmer", 
    formula = paste0(f, " + (1|individual_id)")
)

original_estimates <- associations_me$summaries$estimate[
    associations_me$summaries$term == "case_controlPOSITIVE"
]
replication_estimates <- replication_me$summaries$estimate[
    replication_me$summaries$term == "case_controlPOSITIVE"
]

cor.test(original_estimates, replication_estimates)

## ----module_preservation------------------------------------------------------
# Using a low number of permutations (10) for the vignette
mod_pres_res <- module_preservation(
    wave1_me, 
    wave2_se, 
    verbose = 0, 
    nPermutations = 5
)

plot_module_preservation(mod_pres_res)

## ----extension_vignette, eval=FALSE-------------------------------------------
# vignette("Extensions", package = "SummarizedExperiment")

## ----extend_class-------------------------------------------------------------
PCAExperiment <- setClass(
    "PCAExperiment",
    contains = "ReducedExperiment",
    slots = representation(
        rotation = "matrix",
        sdev = "numeric"
    )
)

## -----------------------------------------------------------------------------
prcomp_res <- stats::prcomp(t(assay(wave1_se)), center = TRUE, scale. = TRUE)

my_pca_experiment <- PCAExperiment(
    wave1_se,
    reduced = prcomp_res$x,
    rotation = prcomp_res$rotation,
    sdev = prcomp_res$sdev,
    center = prcomp_res$center,
    scale = prcomp_res$scale
)

my_pca_experiment

## -----------------------------------------------------------------------------
reduced(my_pca_experiment)[1:5, 1:5]

## -----------------------------------------------------------------------------
setGeneric("rotation", function(x, ...) standardGeneric("rotation"))

setMethod("rotation", "PCAExperiment", function(x, withDimnames = TRUE) {
    return(x@rotation)
})

rotation(my_pca_experiment)[1:5, 1:5]

## -----------------------------------------------------------------------------
setGeneric("rotation<-", function(x, ..., value) standardGeneric("rotation<-"))

setReplaceMethod("rotation", "PCAExperiment", function(x, value) {
    x@rotation <- value
    validObject(x)
    return(x)
})

# Set the value at position [1, 1] to 10 and get
rotation(my_pca_experiment)[1, 1] <- 10
rotation(my_pca_experiment)[1:5, 1:5]

## ----make_fe_from_pca---------------------------------------------------------
# Construct a SummarizedExperiment (if you don't already have one)
expr_mat <- assay(wave1_se)
pheno_data <- colData(wave1_se)
se <- SummarizedExperiment(
    assays = list("normal" = expr_mat),
    colData = pheno_data
)

# Calculate PCA with prcomp
prcomp_res <- stats::prcomp(t(assay(se)), center = TRUE, scale. = TRUE)

# Create a FactorisedExperiment container
fe <- FactorisedExperiment(
    se,
    reduced = prcomp_res$x,
    loadings = prcomp_res$rotation,
    stability = prcomp_res$sdev,
    center = prcomp_res$center,
    scale = prcomp_res$scale
)

fe

## ----session_info, echo=FALSE-------------------------------------------------
sessionInfo()

