% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modules.R
\name{assess_soft_threshold}
\alias{assess_soft_threshold}
\title{Assess soft thresholding power for WGCNA}
\usage{
assess_soft_threshold(
  X,
  assay_name = "normal",
  powerVector = 1:30,
  RsquaredCut = 0.85,
  max_mean_connectivity = 100,
  corType = "pearson",
  networkType = "signed",
  maxBlockSize = 30000,
  verbose = 0,
  ...
)
}
\arguments{
\item{X}{Either a \link[SummarizedExperiment]{SummarizedExperiment} object
or a matrix containing data to be subject to WGCNA. \code{X} should have rows as
features and columns as samples.}

\item{assay_name}{If \code{X} is a
\link[SummarizedExperiment]{SummarizedExperiment}, then this should be the
name of the assay to be subject to WGCNA.}

\item{powerVector}{A vector of integers indicating soft-thresholding powers
to be assessed. See \link[WGCNA]{pickSoftThreshold}.}

\item{RsquaredCut}{The minimum scale free topology fitting index required.
Used to select the soft-thresholding power.
See \link[WGCNA]{pickSoftThreshold}.}

\item{max_mean_connectivity}{The maximal mean connectivity required. Used
to select the soft-thresholding power.}

\item{corType}{The type of correlation to be used to generate a correlation
matrix during network formation. One of "pearson" (\link[WGCNA]{cor}) and
"bicor" (\link[WGCNA]{bicor}).}

\item{networkType}{The type of network to be generated, either "unsigned",
"signed" or "signed hybrid". See \link[WGCNA]{pickSoftThreshold} for
more details.}

\item{maxBlockSize}{The chunk size (in terms of the number of features/genes)
to process the data. See \link[WGCNA]{pickSoftThreshold} for
more details. The default (30000) should process standard transcriptomic
datasets in a single chunk. Results may differ if the number of features
exceeds the chunk size. Lower values of this parameter will use less memory
to calculate networks.}

\item{verbose}{The verbosity, passed to \link[WGCNA]{pickSoftThreshold}.}

\item{...}{Additional arguments to be passed to
\link[WGCNA]{pickSoftThreshold}.}
}
\value{
Returns a \code{data.frame},
generated by \link[WGCNA]{pickSoftThreshold},
with scale free topology fitting indices and connectivity statistics.
Additionally contains a column, \code{estimated_power}, indicating the
recommended power to use (see \code{details}). We suggest manually considering
suitability of the soft-thresholding power rather than solely relying on this
automated approach.
}
\description{
A wrapper around \link[WGCNA]{pickSoftThreshold}, allowing assessment and
automatic selection of soft-thresholding power. Extends the function to
accept a \link[SummarizedExperiment]{SummarizedExperiment} as input and
additionally considers mean connectivity when selecting the
soft-thresholding power to recommend.
}
\details{
The \link[WGCNA]{pickSoftThreshold} function estimates the power by
selecting the lowest value with
a minimum scale free topology fitting index exceeding \code{RsquaredCut}.
The \code{assess_soft_threshold} function mirrors this behaviour when
\code{max_mean_connectivity} is \code{NULL}. When \code{max_mean_connectivity} is
specified, however, we additionally require that the selected power
does not exceed this connectivity threshold.
}
\examples{
# Get the airway data as a SummarizedExperiment (with a subset of features)
set.seed(2)
airway_se <- ReducedExperiment:::.get_airway_data(n_features = 500)

# Select soft-thresholding power to use (use capture.output to hide WGCNA's prints)
WGCNA::disableWGCNAThreads()
invisible(capture.output(fit_indices <- assess_soft_threshold(airway_se)))

print(fit_indices)
print(paste0("Estimated power: ", fit_indices$Power[fit_indices$estimated_power]))

}
\seealso{
\code{\link[WGCNA:pickSoftThreshold]{WGCNA::pickSoftThreshold()}},
\code{\link[=run_wgcna]{run_wgcna()}}
}
\author{
Jack Gisby
}
